
// conviolutionCL.cpp
//-------------------
// Ce programme implante un calcul de convolution sur une image en utilisant la carte graphique (openCL)
// Le filtre utilisé est un filtre moyenneur.
// Le programme ne fonctionne que sur des images de type «raw»
//
// Paramètres sur la lignre de commandes
//  - Parm #1 : nombre de coeurs à utiliser
//  - Parm #2 : nombre de groupes à utiliser (mettre 0 généralement)
//  - Parm #3 : taille «horizontale» de l'image 
//  - Parm #4 : taille «verticale» de l'image 
//  - Parm #5 : nom du fichier contenant l'image
//  - Parm #6 : nom du fichier de sortie
//  - Parm #7 : taille du filtre

#include <stdlib.h>
#include <stdio.h>
#include <CL/cl.hpp>
#include <CL/cl.h>
#include <iostream>
#include <sstream>
#include <chrono>
#include <fstream> 
#include <vector>

using namespace std;
using namespace std::chrono;

int main(int argc, char *argv[])
{
    int taille_filtre;
    int th_image, tv_image;
    ifstream image;
    ofstream image_res;
    string nom,nom_sortie, tmp;
    const char* texte =0; 
    int nbCoeurs, nbGroupes;
    std::ifstream fichierSource;         // Pour le fichier source du noyau/

    // Variables nécessaire pour l'environnement OpenCL
    std::vector<cl::Platform> mes_fabricants;  // Marque des cartes
    std::vector<cl::Device> mes_gpu;           // Liste des GPU
    cl::Platform fabricants_par_defaut ;       // Le fabricant choisi
    cl::Device gpu_par_defaut;                 // La carte choisie
    cl::Context contexte;                      // Contexte nécessaire pour OpenCL
    cl::CommandQueue listeCmd;                 // Liste de commandes à exécuter sur la carte
    cl_int err; 
    cl::Kernel noyau;                          // Le noyau à exécuter
    cl::Buffer param[2];                       // Les tampons mémoire sur la carte

 
    // Récupération des paramètres
    nbCoeurs = stoi(argv[1]);
    nbGroupes = stoi(argv[2]);
    th_image = stoi(argv[3]);
    tv_image = stoi(argv[4]);   
    nom = argv[5];
    nom_sortie = argv[6];
    taille_filtre = stoi(argv[7]);
    // **** Pour mise au point *****
    //cout << "Convolution sur " << nbCoeurs << " coeurs de  " << nom << " de taille " << th_image << "x" 
    //     << tv_image << " (filtre = " << taille_filtre << ")"<< endl;
    //---------------------------------------------------------------
    // Lecture de l'image dans un tableau de caratères (à une dimension)
    image.open(nom);         // ouverture du fichier
    if (image.is_open())
    {
      ostringstream ss;      // on veut lire dans une stream
      ss << image.rdbuf();   // lecture de tout le fichier (fichier raw seulement)
      tmp = ss.str();        // on convertit en stribg
      texte = tmp.c_str();   // on place l'image dans une chaine de caractères
      image.close();
    }
    else cout << "Impossible d'ouvrir le fichier"; 
    int taille_texte = tmp.size();  // récupération de la taille de l'aimge (en 1D)
    // ------
    /// À des fins de mise au point
    // cout << "taille du texte = " << taille_texte << endl;
    // ------
    // On alloue le tableau pour l'image de sortie
    char * sortie = new char[taille_texte];  
    //-------
    // On calcule le filtre moyenneur
    double valeur = 1.0/ (taille_filtre * taille_filtre);
    //-------
    // On détermine la dimension de l'image à traiter par chaque fil (ou coeurs)
    // On ignore le reste si la taille de l'image ne se divise pas par le nombre de fils
    int temp;
    temp = taille_texte / nbCoeurs;           // Chacun traite un fraction de l'image
    // -----
    // On positionne le chronomètre pour évaluer la performance
     auto temps_depart = system_clock::now();

    //===================================================================
    // Début initialisation opencl
    // ----
    // On obtient les marques des cartes
    cl::Platform::get(&mes_fabricants);
    if (mes_fabricants.size() == 0) 
    {
        std::cerr << "Error: Failed to find a platform!" << std::endl;
        return EXIT_FAILURE;
    }
    //----
    // Choix de la carte «0» comme carte par défaut (à changer selon votre environnement
    fabricants_par_defaut = mes_fabricants[0];
    std::cout << "fabricants par defaut "<<fabricants_par_defaut.getInfo<CL_PLATFORM_NAME>()<< std::endl;  
    // ----
    // Création du contexte avec la marque par défaut...
    cl_context_properties properties[] = { CL_CONTEXT_PLATFORM, (cl_context_properties)(fabricants_par_defaut)(), 0};
    contexte = cl::Context(CL_DEVICE_TYPE_GPU, properties);
    // -----
    // Obtention de l'information sur la carte à utiliser
    mes_gpu = contexte.getInfo<CL_CONTEXT_DEVICES>();
    std::cout << "Nombre de cartes GPU " << mes_gpu.size()<< std::endl;
    // Utilisation de la carte «0» -- à changer selon votre environnement
    gpu_par_defaut=mes_gpu[0];
    std::cout<< "Utilise par defaut gpu : "<<gpu_par_defaut.getInfo<CL_DEVICE_NAME>()<<std::endl ;
    // ----
    // Création de la liste de commande pour la carte par défaut
    listeCmd = cl::CommandQueue(contexte, gpu_par_defaut, 0, &err); cout << "Liste de commandes "<< err << endl;
    // ----
    // Obtention de l'information sur la carte (Optionnel)
    // int num;
    // mes_gpu[0].getInfo(CL_DEVICE_MAX_COMPUTE_UNITS, &num);
    // cout << "Nombre d'unités de calcul : " << num << endl;
    // mes_gpu[0].getInfo(CL_DEVICE_MAX_WORK_ITEM_SIZES, &num);
    // cout << "Taille des work item : " << num << endl;

    // std::cout << "\t\tDevice Name: " << gpu_par_defaut.getInfo<CL_DEVICE_NAME>() << std::endl;  
    // std::cout << "\t\tDevice Type: " << gpu_par_defaut.getInfo<CL_DEVICE_TYPE>();
    // std::cout << " (GPU: " << CL_DEVICE_TYPE_GPU << ", CPU: " << CL_DEVICE_TYPE_CPU << ")" << std::endl; 
    // std::cout << "\t\tDevice Vendor: " << gpu_par_defaut.getInfo<CL_DEVICE_VENDOR>() << std::endl;
    // std::cout << "\t\tDevice Max Compute Units: " << gpu_par_defaut.getInfo<CL_DEVICE_MAX_COMPUTE_UNITS>() << std::endl;
    // std::cout << "\t\tDevice Global Memory: " << gpu_par_defaut.getInfo<CL_DEVICE_GLOBAL_MEM_SIZE>() << std::endl;
    // std::cout << "\t\tDevice Max Clock Frequency: " << gpu_par_defaut.getInfo<CL_DEVICE_MAX_CLOCK_FREQUENCY>() << std::endl;
    // std::cout << "\t\tDevice Max Allocateable Memory: " << gpu_par_defaut.getInfo<CL_DEVICE_MAX_MEM_ALLOC_SIZE>() << std::endl;
    // std::cout << "\t\tDevice Local Memory: " << gpu_par_defaut.getInfo<CL_DEVICE_LOCAL_MEM_SIZE>() << std::endl;
    // std::cout << "\t\tDevice Available: " << gpu_par_defaut.getInfo< CL_DEVICE_AVAILABLE>() << std::endl;
    //-----------------------------------------------------------------------
    // Lecture et compilation du noyau
    //-----------------------------------------------------------------------
	fichierSource.open("./kernel.cl");
    /* Chargement du code source du noyau */
    std::string code((std::istreambuf_iterator<char>(fichierSource)), std::istreambuf_iterator<char>());
	cl::Program::Sources source(1, std::make_pair(code.c_str(), code.length()+1));
    // On associe le programme au contexte
    cl::Program program = cl::Program(contexte, source);
    // On compile le programme selon le contexte
    err = program.build(contexte.getInfo<CL_CONTEXT_DEVICES>()); cout << "Erreur (compilation) = " << err << endl;
    std::cout << "program.build: " <<  err << std::endl;
     if (err != CL_SUCCESS){
        std::cout << "Build Status: " << program.getBuildInfo<CL_PROGRAM_BUILD_STATUS>(gpu_par_defaut) << std::endl;
        std::cout << "Build Options:\t" << program.getBuildInfo<CL_PROGRAM_BUILD_OPTIONS>(gpu_par_defaut) << std::endl;
       std::cout << "Build Log:\t " << program.getBuildInfo<CL_PROGRAM_BUILD_LOG>(gpu_par_defaut) << std::endl;
    }
    // Création du noyau exécutable
    noyau = cl::Kernel(program, "convo", &err);cout << "Erreur (création du noyau) = " << err << endl;
    //-----------------------------------------------------------------------
    // Allocation de la mémoire pour les images d'entrée et de sortie
    //-----------------------------------------------------------------------
    param[0] = cl::Buffer(contexte, CL_MEM_READ_WRITE, taille_texte * sizeof(char), NULL, &err);
    cout << "Erreur (buf) = " << err << endl;
    std::cout << "buffer = " << err << " , taille = " << taille_texte * sizeof(char)<< std::endl;
    param[1] = cl::Buffer(contexte, CL_MEM_READ_ONLY, taille_texte * sizeof(char), NULL, &err);
    cout << "Erreur (buf) = " << err << endl;
    std::cout << "buffer = " << err << " , taille = " << taille_texte * sizeof(char)<< std::endl;
    // Transfert de l'image vers a carte
     err = listeCmd.enqueueWriteBuffer(param[1], CL_TRUE, 0, taille_texte*sizeof(char), texte);
    //-----------------------------------------------------------------------
    // Initialisation des parammètres du noyau
    //-----------------------------------------------------------------------
    err = noyau.setArg(0,param[0]); cout << "Erreur (setarg) = " << err << endl;
    err = noyau.setArg(1,param[1]); cout << "Erreur (setarg) = " << err << endl;
    err = noyau.setArg(2,temp);     cout << "Erreur (setarg) = " << err << endl;
    err = noyau.setArg(3,valeur);   cout << "Erreur (setarg) = " << err << endl;
    err = noyau.setArg(4,taille_filtre); cout << "Erreur (setarg) = " << err << endl;
    err = noyau.setArg(5,th_image); cout << "Erreur (setarg) = " << err << endl;
    err = noyau.setArg(6,tv_image); cout << "Erreur (setarg) = " << err << endl;
    //-----------------------------------------------------------------------
    // Lancement de l'exécution
    //-----------------------------------------------------------------------
    if (nbGroupes == 0) 
    {
        cout<< "début exécution du noyau (nbGroupe=0)"<<endl;
        err = listeCmd.enqueueNDRangeKernel(noyau, cl::NullRange, cl::NDRange(nbCoeurs), cl::NullRange);
        
    }
    else err = listeCmd.enqueueNDRangeKernel(noyau, cl::NullRange, cl::NDRange(nbCoeurs), cl::NDRange(nbCoeurs/nbGroupes)); 
    // Attente de la fin de l'exécution
    listeCmd.finish(); std::cout << "terminé.......=============" << std::endl;
    //-----------------------------------------------------------------------
    // Récupération des résultats
    //-----------------------------------------------------------------------
    err = listeCmd.enqueueReadBuffer(param[0], CL_TRUE, 0, taille_texte, sortie);  cout << "Erreur (resul) = " << err << endl;

   //===============================================
    // On arrête le chronomètre
    auto temps_arret = system_clock::now();
    //----------------
    // On sauve le résultat dans un fichier 
    image_res.open(nom_sortie,ios::binary | ios::out);
    if (image_res.is_open())
    {
      for (int ind=0; ind < taille_texte; ind++)
        image_res << sortie[ind];
      image_res.close();
    }
    else cout << "Impossible d'ouvrir le fichier"; 
    // ----------------
    // Affichage u temps d'exécution
    auto duree = duration_cast<milliseconds>(temps_arret - temps_depart);
    cout << "Temps execution = " << duree.count()/1000.0 << " secondes" << endl;
    return 0;
}
